from collections import OrderedDict

import numpy as np

from sfsimodels.models.abstract_models import PhysicalObject
from sfsimodels.models import SeismicHazard, Foundation, Soil
from sfsimodels.models.material import Concrete
from sfsimodels.exceptions import ModelError, deprecation
from sfsimodels import functions as sf


class Building(PhysicalObject):
    """
    An object to define Buildings
    """
    _id = None
    name = None
    base_type = "building"
    type = "building"
    _floor_length = None
    _floor_width = None
    _interstorey_heights = np.array([0.0])  # m
    _storey_masses = np.array([0.0])  # kg
    _concrete = Concrete()
    _n_storeys = None
    _g = 9.81  # m/s2  # gravity

    def __init__(self, n_storeys, verbose=0, **kwargs):
        print(**kwargs)
        super(Building, self).__init__()
        self._n_storeys = n_storeys
        if not hasattr(self, "inputs"):
            self.inputs = []
        self._extra_class_variables = [
                "id",
                "name",
                "type",
                'floor_length',
                'floor_width',
                'interstorey_heights',
                'storey_masses'
            ]
        self.inputs += self._extra_class_variables
        self.all_parameters = self.inputs + [
            "n_storeys"
        ]

    def to_dict(self, extra=()):
        outputs = OrderedDict()
        skip_list = []
        full_inputs = self.inputs + list(extra)
        for item in full_inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                outputs[item] = sf.collect_serial_value(value)
        return outputs

    @property
    def ancestor_types(self):
        return super(Building, self).ancestor_types + ["building"]

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def floor_length(self):
        return self._floor_length

    @floor_length.setter
    def floor_length(self, value):
        self._floor_length = value

    @property
    def floor_width(self):
        return self._floor_width

    @floor_width.setter
    def floor_width(self, value):
        self._floor_width = value

    @property
    def g(self):
        return self._g

    @property
    def concrete(self):
        return self._concrete

    @concrete.setter
    def concrete(self, conc_inst):
        self._concrete = conc_inst

    @property
    def floor_area(self):
        try:
            return self.floor_length * self.floor_width
        except TypeError:
            return None

    @property
    def heights(self):
        return np.cumsum(self._interstorey_heights)

    @property
    def max_height(self):
        return np.sum(self._interstorey_heights)

    @property
    def n_storeys(self):
        return self._n_storeys

    @property
    def interstorey_heights(self):
        return self._interstorey_heights

    @interstorey_heights.setter
    def interstorey_heights(self, heights):
        if len(heights) != self.n_storeys:
            raise ModelError("Specified heights must match number of storeys (%i)." % self.n_storeys)
        self._interstorey_heights = np.array(heights)

    @property
    def storey_masses(self):
        return self._storey_masses

    @storey_masses.setter
    def storey_masses(self, masses):
        self._storey_masses = np.array(masses)

    def set_storey_masses_by_pressure(self, stresses):
        if hasattr(stresses, "length"):
            if len(stresses) != self.n_storeys:
                raise ModelError("Length of defined storey pressures: {0}, "
                                 "must equal number of stories: {1}".format(len(stresses), self.n_storeys))
            self.storey_masses = stresses * self.floor_area / self._g
        else:
            self.storey_masses = stresses * np.ones(self.n_storeys) * self.floor_area / self._g


class Section(PhysicalObject):  # not used?
    id = None
    type = "section"
    base_type = "section"
    _depth = None
    _width = None

    def __init__(self):
        self.inputs = ["depth",
                       "width"]

    def to_dict(self, extra=()):
        outputs = OrderedDict()
        skip_list = []
        full_inputs = self.inputs + list(extra)
        for item in full_inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                outputs[item] = sf.collect_serial_value(value)
        return outputs

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        self._depth = value

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = value


class Element(PhysicalObject):
    section_lengths = None

    def __init__(self):
        self.sections = [Section()]

    @property
    def s(self):
        return self.sections

    def set_section_prop(self, prop, prop_value, sections=None):
        if sections is None:
            sections = range(len(self.sections))
        for sect_i in sections:
            setattr(self.sections[sect_i], prop, prop_value)

    def add_inputs_to_section(self, props, sections=None):
        if sections is None:
            sections = range(len(self.sections))
        for sect_i in sections:
            self.sections[sect_i].inputs += props

    def split_into_multiple(self, lengths):
        section = self.sections[0]
        self.section_lengths = []
        self.sections = []
        for length in lengths:
            self.section_lengths.append(length)
            self.sections.append(section.deepcopy())

    def get_section_prop(self, prop, section_i=0):
        return getattr(self.sections[section_i], prop)

    def to_dict(self, extra=()):
        output = []
        for i in range(len(self.sections)):
            output.append(self.sections[i].to_dict(extra=extra))
        return []


class Frame(object):
    _bay_lengths = None

    def __init__(self, n_storeys, n_bays):
        if not hasattr(self, "inputs"):
            self.inputs = []
        self._extra_class_variables = [
            "beams",
            "columns",
            "bay_lengths"
        ]
        self.inputs += self._extra_class_variables
        self._n_storeys = n_storeys
        self._n_bays = n_bays
        self._allocate_beams_and_columns()

    def to_dict(self, extra=()):
        outputs = OrderedDict()
        skip_list = []
        # skip_list = ["beams", "columns"]  # TODO: uncomment this
        # skip_list = ["beams"]
        full_inputs = self.inputs + list(extra)
        for item in full_inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                outputs[item] = sf.collect_serial_value(value)

        # # Deal with sections
        # beam_sections = OrderedDict()
        # beam_section_ids = []
        # beam_section_count = 1
        # for ss in range(self.n_storeys):
        #     beam_section_ids.append([])
        #     for bb in range(self.n_bays):
        #         beam_section_ids.append(beam_section_count)
        #         # build a hash string of the section inputs to check uniqueness
        #         parts = []
        #         for item in self.beams[ss][bb].inputs:
        #             parts.append(item)
        #             parts.append(getattr(self.beams[ss][bb], item))
        #         p_str = "-".join(parts)
        #         if p_str not in beam_sections:
        #             beam_sections[p_str] = self.beams[ss][bb].to_dict(extra)
        #             beam_sections[p_str]["id"] = beam_section_count
        #             beam_section_count += 1
        #
        # outputs["beam_section_ids"] = beam_section_ids
        # outputs["beam_sections"] = OrderedDict()
        # for i, p_str in enumerate(beam_sections):
        #     outputs["beam_sections"][beam_sections[p_str]["id"]] = beam_sections[p_str]
        return outputs

    @property
    def ancestor_types(self):
        return ["frame"]

    def _allocate_beams_and_columns(self):
        self._beams = [[Element() for i in range(self.n_bays)] for ss in range(self.n_storeys)]
        self._columns = [[Element() for i in range(self.n_cols)] for ss in range(self.n_storeys)]

    @property
    def beams(self):
        return self._beams

    @property
    def columns(self):
        return self._columns

    @property
    def n_bays(self):
        return self._n_bays

    @property
    def n_storeys(self):
        return self._n_storeys

    @property
    def n_cols(self):
        return self._n_bays + 1

    @n_bays.setter
    def n_bays(self, value):
        raise ModelError("Can not set n_bays, only on initialisation")

    # @column_sections.setter
    # def column_sections(self, columns: dict):
    #     assert len(columns) ==

    def set_beam_prop(self, prop, values, repeat="up"):
        """
        Specify the properties of the beams

        :param values:
        :param repeat: if 'up' then duplicate up the structure
        :return:
        """
        values = np.array(values)
        if repeat == "up":
            assert len(values.shape) == 1
            values = [values for ss in range(self.n_storeys)]
        else:
            assert len(values.shape) == 2
        if len(values[0]) != self.n_bays:
            raise ModelError("beam depths does not match number of bays (%i)." % self.n_bays)
        for ss in range(self.n_storeys):
            for i in range(self.n_bays):
                self._beams[ss][i].set_section_prop(prop, values[0][i])

    def set_column_prop(self, prop, values, repeat="up"):
        """
        Specify the properties of the columns

        :param values:
        :param repeat: if 'up' then duplicate up the structure
        :return:
        """
        values = np.array(values)
        if repeat == "up":
            assert len(values.shape) == 1
            values = [values for ss in range(self.n_storeys)]
        else:
            assert len(values.shape) == 2
        if len(values[0]) != self.n_cols:
            raise ModelError("column props does not match n_cols (%i)." % self.n_cols)
        for ss in range(self.n_storeys):
            for i in range(self.n_cols):
                self._columns[ss][i].set_section_prop(prop, values[0][i])

    def beams_at_storey(self, storey):
        return self._beams[storey - 1]

    @property
    def beam_depths(self):
        beam_depths = []
        for ss in range(self.n_storeys):
            beam_depths.append([])
            for i in range(self.n_bays):
                beam_depths[ss].append(self.beams[ss][i].get_section_prop("depth"))
        return np.array(beam_depths)

    @property
    def beam_widths(self):
        beam_widths = []
        for ss in range(self.n_storeys):
            beam_widths.append([])
            for i in range(self.n_bays):
                beam_widths[ss].append(self.beams[ss][i].get_section_prop("width"))
        return np.array(beam_widths)

    @property
    def column_depths(self):
        column_depths = []
        for ss in range(self.n_storeys):
            column_depths.append([])
            for i in range(self.n_cols):
                column_depths[ss].append(self.columns[ss][i].get_section_prop("depth"))
        return np.array(column_depths)

    @property
    def column_widths(self):
        column_widths = []
        for ss in range(self.n_storeys):
            column_widths.append([])
            for i in range(self.n_cols):
                column_widths[ss].append(self.columns[ss][i].get_section_prop("width"))
        return np.array(column_widths)

    @property
    def bay_lengths(self):
        return self._bay_lengths

    @bay_lengths.setter
    def bay_lengths(self, bay_lengths):
        if len(bay_lengths) != self.n_bays:
            raise ModelError("bay_lengths does not match number of bays (%i)." % self.n_bays)
        self._bay_lengths = np.array(bay_lengths)


class BuildingFrame(Frame, Building):
    _n_seismic_frames = None
    _n_gravity_frames = None
    type = "building_frame"

    def __init__(self, n_storeys, n_bays):
        Frame.__init__(self, n_storeys, n_bays)
        Building.__init__(self, n_storeys)
        # super(BuildingFrame, self).__init__(n_storeys, n_bays)  # run parent class initialiser function
        self._extra_class_inputs = ["n_seismic_frames",
                           "n_gravity_frames"]
        self.inputs = self.inputs + self._extra_class_inputs
        # Frame.__init__(self, n_storeys, n_bays)
        # Building.__init__(self, n_storeys, n_bays)

    @property
    def ancestor_types(self):
        return super(BuildingFrame, self).ancestor_types + ["building_frame"]

    @property
    def n_seismic_frames(self):
        return self._n_seismic_frames

    @n_seismic_frames.setter
    def n_seismic_frames(self, value):
        self._n_seismic_frames = value

    @property
    def n_gravity_frames(self):
        return self._n_gravity_frames

    @n_gravity_frames.setter
    def n_gravity_frames(self, value):
        self._n_gravity_frames = value


class FrameBuilding(BuildingFrame):
    def __init__(self, n_storeys, n_bays):
        deprecation("FrameBuilding is deprecated, use BuildingFrame.")
        super(FrameBuilding, self).__init__(n_storeys, n_bays)


class BuildingFrame2D(Frame, Building):
    _extra_class_inputs = []
    type = "building_frame2D"

    def __init__(self, n_storeys, n_bays):
        Frame.__init__(self, n_storeys, n_bays)
        Building.__init__(self, n_storeys)
        # super(FrameBuilding2D, self).__init__(n_storeys, n_bays)  # run parent class initialiser function
        self.inputs = self.inputs + self._extra_class_inputs

    @property
    def ancestor_types(self):
        return ["physical_object", "frame", "building"] + ["building_frame2D"]  # TODO: improve this logic

    # def to_dict(self, extra=()):
    #     outputs = OrderedDict()
    #     skip_list = []
    #     full_inputs = self.inputs + list(extra)
    #     for item in full_inputs:
    #         if item not in skip_list:
    #             value = self.__getattribute__(item)
    #             outputs[item] = sf.collect_serial_value(value)
    #     return outputs

    def to_dict(self, extra=()):
        outputs = OrderedDict()
        skip_list = ["beams", "columns"]
        full_inputs = self.inputs + list(extra)
        for item in full_inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                outputs[item] = sf.collect_serial_value(value)

        # Deal with sections
        column_sections = OrderedDict()
        column_section_ids = []
        column_section_count = 0
        for ss in range(self.n_storeys):
            column_section_ids.append([])
            for cc in range(self.n_cols):
                # build a hash string of the section inputs to check uniqueness
                parts = []
                column_section_ids[ss].append([])
                for sect_i in range(len(self.columns[ss][cc].sections)):
                    for item in self.columns[ss][cc].sections[sect_i].inputs:
                        parts.append(item)
                        parts.append(str(self.columns[ss][cc].get_section_prop(item, section_i=sect_i)))
                    p_str = "-".join(parts)
                    if p_str not in column_sections:
                        column_sections[p_str] = self.columns[ss][cc].sections[sect_i].to_dict(extra)
                        column_section_count += 1
                        column_sections[p_str]["id"] = column_section_count
                    column_section_ids[ss][cc].append(column_section_count)

        beam_sections = OrderedDict()
        beam_section_ids = []
        beam_section_count = 0
        for ss in range(self.n_storeys):
            beam_section_ids.append([])
            for bb in range(self.n_bays):
                # build a hash string of the section inputs to check uniqueness
                parts = []
                beam_section_ids[ss].append([])
                for sect_i in range(len(self.beams[ss][bb].sections)):
                    for item in self.beams[ss][bb].sections[sect_i].inputs:
                        parts.append(item)
                        parts.append(str(self.beams[ss][bb].get_section_prop(item, section_i=sect_i)))
                    p_str = "-".join(parts)
                    if p_str not in beam_sections:
                        beam_sections[p_str] = self.beams[ss][bb].sections[sect_i].to_dict(extra)
                        beam_section_count += 1
                        beam_sections[p_str]["id"] = beam_section_count
                    beam_section_ids[ss][bb].append(beam_section_count)

        outputs["column_section_ids"] = column_section_ids
        outputs["beam_section_ids"] = beam_section_ids
        outputs["column_sections"] = OrderedDict()
        outputs["beam_sections"] = OrderedDict()
        for i, p_str in enumerate(column_sections):
            outputs["column_sections"][column_sections[p_str]["id"]] = column_sections[p_str]
        for i, p_str in enumerate(beam_sections):
            outputs["beam_sections"][beam_sections[p_str]["id"]] = beam_sections[p_str]
        return outputs


class FrameBuilding2D(BuildingFrame2D):
    def __init__(self, n_storeys, n_bays):
        deprecation("FrameBuilding2D class is deprecated, use BuildingFrame2D.")
        super(FrameBuilding2D, self).__init__(n_storeys, n_bays)


class BuildingWall(Building):  # new name
    n_walls = 1
    wall_depth = 0.0  # m
    wall_width = 0.0  # m
    type = "building_wall"

    def __init__(self, n_storeys):
        super(BuildingWall, self).__init__(n_storeys)  # run parent class initialiser function
        self._extra_class_inputs = [
            "n_walls",
            "wall_depth",
            "wall_width"
        ]
        self.inputs += self._extra_class_inputs

    @property
    def ancestor_types(self):
        return super(BuildingWall, self).ancestor_types + ["building_wall"]


class WallBuilding(BuildingWall):
    def __init__(self, n_storeys):
        deprecation("WallBuilding class is deprecated, use BuildingWall.")
        super(WallBuilding, self).__init__(n_storeys)


class BuildingSDOF(PhysicalObject):
    """
    An object to describe structures.
    """
    _id = None
    name = None
    base_type = "building"
    type = "sdof"
    _h_eff = None
    _mass_eff = None
    _t_fixed = None
    _mass_ratio = None

    def __init__(self, g=9.8):
        self.inputs = [
            "id",
            "name",
            "base_type",
            "type",
            "h_eff",
            "mass_eff",
            "t_fixed",
            "mass_ratio"
        ]
        self._g = g

    def to_dict(self):
        outputs = OrderedDict()
        skip_list = []
        for item in self.inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                if isinstance(value, int):
                    outputs[item] = str(value)
                else:
                    outputs[item] = value
        return outputs

    @property
    def ancestor_types(self):
        return super(BuildingSDOF, self).ancestor_types + ["sdof"]

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = int(value)

    @property
    def h_eff(self):
        return self._h_eff

    @h_eff.setter
    def h_eff(self, value):
        if value is None or value == "":
            return
        self._h_eff = float(value)

    @property
    def mass_eff(self):
        return self._mass_eff

    @mass_eff.setter
    def mass_eff(self, value):
        if value is None or value == "":
            return
        self._mass_eff = float(value)

    @property
    def t_fixed(self):
        return self._t_fixed

    @t_fixed.setter
    def t_fixed(self, value):
        if value is None or value == "":
            return
        self._t_fixed = float(value)

    @property
    def mass_ratio(self):
        return self._mass_ratio

    @mass_ratio.setter
    def mass_ratio(self, value):
        if value is None or value == "":
            return
        self._mass_ratio = float(value)

    @property
    def k_eff(self):
        return 4.0 * np.pi ** 2 * self.mass_eff / self.t_fixed ** 2

    @property
    def weight(self):
        return self.mass_eff / self.mass_ratio * self._g


class Structure(BuildingSDOF):
    def __init__(self, g=9.8):
        deprecation("Structure class is deprecated, use BuildingSDOF.")
        super(Structure, self).__init__(g=g)


class SoilStructureSystem(PhysicalObject):
    bd = BuildingSDOF()
    fd = Foundation()
    sp = Soil()
    hz = SeismicHazard()
    name = "Nameless"

    inputs = ["name"] + bd.inputs + fd.inputs + sp.inputs + hz.inputs


if __name__ == '__main__':
    print(BuildingFrame2D.__mro__)
    a = BuildingFrame2D(1, 2)
    print(a.n_bays)



from sfsimodels.models.hazards import SeismicHazard
from sfsimodels.models.foundations import Foundation, FoundationPad, FoundationRaft
from sfsimodels.models.soils import Soil, SoilCritical, SoilProfile
from sfsimodels.models.buildings import Building, BuildingFrame, BuildingWall, BuildingSDOF, BuildingFrame2D, Section
from sfsimodels.models import material
from sfsimodels.models.systems import SoilStructureSystem

from sfsimodels.models.buildings import FrameBuilding, WallBuilding, Structure, FrameBuilding2D  # deprecated objects
from sfsimodels.models.foundations import PadFoundation, RaftFoundation  # deprecated objects



from sfsimodels import models
from sfsimodels import properties
from sfsimodels import exceptions
from collections import OrderedDict
vp = properties.pp


def build_parameter_descriptions(obj, user_p=None, output="csv", show_none=True, ignore=None, plist=None):
    """
    Creates a list of the decription of all the inputs of an object

    :param obj: object, that has parameters
    :param user_p: dict, user defined parameter descriptions
    :param show_none: if false, only shows descriptions of parameters that are not None
    :param ignore: list of parameters to not build
    :return:
    """
    if user_p is None:
        user_p = {}
    if ignore is None:
        ignore = []
    para = [[obj.__class__.__name__ + " inputs:", "", ""]]
    if plist is None:
        if not hasattr(obj, 'inputs'):
            raise exceptions.ModelError("Object must contain parameter 'inputs' or set plist as an input")
        plist = obj.inputs

    p_dict = {}
    if hasattr(obj, 'ancestor_types'):
        bt = obj.ancestor_types
        for otype in bt:  # cycles from lowest class, so descriptions get overridden
            if otype in vp:
                for item in vp[otype]:
                    p_dict[item] = vp[otype][item]
        for item in user_p:  # user defined definitions override base ones
            p_dict[item] = user_p[item]
    else:
        p_dict = user_p

    for item in plist:
        if show_none is False and getattr(obj, item) is None:
            continue
        if item in ignore:
            continue
        if item in p_dict:
            para.append([item, p_dict[item][1], p_dict[item][0]])
        else:
            para.append([item, "", ""])
    if output == "csv":
        out_obj = []
        for i in range(len(para)):
            out_obj.append(",".join(para[i]))
    elif output == "list":
        out_obj = para
    elif output == "dict":
        out_obj = OrderedDict()
        for i in range(len(para)):
            out_obj[para[i][0]] = {"description": para[i][2], "units": para[i][1]}
    else:
        raise ValueError("output must be either: 'csv', 'dict' or 'list'.")
    return out_obj


def all_descriptions():
    """
    Generates a list of descriptions of all the models

    :return:
    """
    para = []
    para += build_parameter_descriptions(models.Soil()) + [",,\n"]
    para += build_parameter_descriptions(models.SoilProfile()) + [",,\n"]
    para += build_parameter_descriptions(models.Foundation()) + [",,\n"]
    para += build_parameter_descriptions(models.FoundationPad()) + [",,\n"]
    para += build_parameter_descriptions(models.BuildingSDOF()) + [",,\n"]
    para += build_parameter_descriptions(models.BuildingFrame2D(1, 1))

    return para


if __name__ == '__main__':
    # all_descriptions()
    para = []
    para += build_parameter_descriptions(models.Soil()) + [",,\n"]
    para += build_parameter_descriptions(models.SoilProfile()) + [",,\n"]
    para += build_parameter_descriptions(models.Foundation()) + [",,\n"]
    para += build_parameter_descriptions(models.PadFoundation()) + [",,\n"]
    para += build_parameter_descriptions(models.Structure()) + [",,\n"]
    para += build_parameter_descriptions(models.BuildingFrame2D(1, 1))
    print("\n".join(para))



from collections import OrderedDict
import copy
# from sfsimodels.loader import add_inputs_to_object
import types
from sfsimodels.exceptions import ModelError


class PhysicalObject(object):
    _counter = 0
    type = "physical_object"

    def __iter__(self):  # real signature unknown
        return self

    # def __init__(self, **kwargs):
    #     # super(PhysicalObject, self).__init__()
    #     print("Initialised")

    @property
    def attributes(self):
        all_attributes = []
        for item in self.__dir__():
            if item in ["deepcopy", "set", "to_dict"]:
                continue
            if isinstance(item, types.MethodType):
                continue
            if "_" != item[0]:
                all_attributes.append(item)
        all_attributes.sort()
        return all_attributes

    def __next__(self):
        self._counter += 1
        all_attributes = self.attributes
        if self._counter == len(all_attributes):
            raise StopIteration
        return all_attributes[self._counter]

    def set(self, values):
        """
        Set the object parameters using a dictionary
        """
        if hasattr(self, "inputs"):
            for item in self.inputs:
                if hasattr(self, item):
                    setattr(self, item, values[item])

    def deepcopy(self):
        """ Make a clone of the object """
        return copy.deepcopy(self)

    @property
    def ancestor_types(self):
        return ["physical_object"]

    def add_from_same(self, obj, inputs_from="obj", update_inputs=True):
        if not hasattr(self, "inputs"):
            raise ModelError("self does not contain attribute: 'inputs'")
        if inputs_from == "obj":
            if hasattr(obj, "inputs"):
                inputs_list = obj.inputs
            else:
                raise ModelError("obj does not contain attribute: 'inputs'")
        else:
            inputs_list = self.inputs
        for item in inputs_list:
            if hasattr(obj, item):
                setattr(self, item, getattr(obj, item))
                if update_inputs and item not in self.inputs:
                    self.inputs.append(item)


class CustomObject(PhysicalObject):
    """
    An object to describe structures.
    """
    _id = None
    name = None
    base_type = "custom_object"
    type = "custom_object"

    def __init__(self):
        self.inputs = [
            "id",
            "name",
            "base_type",
            "type"
            ]

    def to_dict(self):
        outputs = OrderedDict()
        skip_list = []
        for item in self.inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                if isinstance(value, int):
                    outputs[item] = str(value)
                else:
                    outputs[item] = value
        return outputs

    @property
    def id(self):
        """
        Object id
        :return:
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def ancestor_types(self):
        return ["custom"]



from collections import OrderedDict
from sfsimodels.models import SoilProfile, Foundation, Building, BuildingSDOF
from sfsimodels.exceptions import ModelError
from sfsimodels import functions as sf


# class SoilSystem(object):
#     id = None
#     name = None
#     type = "soil_system"
#     _sp = SoilProfile()
#     _soil_profile_id = None
#
#     def __init__(self):
#         self.inputs = ["id", "name", "soil_profile_id"]
#
#     def to_dict(self):
#         outputs = OrderedDict()
#         skip_list = []
#         for item in self.inputs:
#             if item not in skip_list:
#                 value = self.__getattribute__(item)
#                 if "_id" in item and value is None:
#                     raise ModelError("Cannot export system with %s set to None" % item)
#                 if item not in skip_list:
#                     value = self.__getattribute__(item)
#                     outputs[item] = sf.collect_serial_value(value)
#         return outputs
#
#     def add_obj_to_system(self, obj):
#         if hasattr(obj, "base_type"):
#             if obj.base_type == "soil":
#                 self.sp = obj
#                 self._soil_profile_id = obj.id
#
#             elif obj.base_type == "foundation":
#                 self.fd = obj
#                 self._foundation_id = obj.id
#             elif obj.base_type == "building":
#                 self.bd = obj
#                 self._building_id = obj.id
#
#     @property
#     def soil_profile_id(self):
#         if self._soil_profile_id is None:
#             if self._sp.id is not None:
#                 self._soil_profile_id = self._sp.id
#         return self._soil_profile_id


class SoilStructureSystem(object):
    id = None
    name = None
    base_type = "system"
    type = "sfs"
    _sp = SoilProfile()
    _bd = BuildingSDOF()
    _fd = Foundation()
    _soil_profile_id = None
    _building_id = None
    _foundation_id = None

    def __init__(self):
        self.inputs = ["id", "name", "soil_profile_id", "building_id", "foundation_id"]

    def to_dict(self):
        outputs = OrderedDict()
        skip_list = []
        for item in self.inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                if "_id" in item and value is None:
                    raise ModelError("Cannot export system with %s set to None" % item)
                if item not in skip_list:
                    value = self.__getattribute__(item)
                    outputs[item] = sf.collect_serial_value(value)
        return outputs

    def add_obj_to_system(self, obj):
        if hasattr(obj, "base_type"):
            if obj.base_type == "soil":
                self.sp = obj
                self._soil_profile_id = obj.id

            elif obj.base_type == "foundation":
                self.fd = obj
                self._foundation_id = obj.id
            elif obj.base_type == "building":
                self.bd = obj
                self._building_id = obj.id

    @property
    def soil_profile_id(self):
        if self._soil_profile_id is None:
            if self._sp.id is not None:
                self._soil_profile_id = self._sp.id
        return self._soil_profile_id

    @property
    def foundation_id(self):
        if self._foundation_id is None:
            if self._fd.id is not None:
                self._foundation_id = self._fd.id
        return self._foundation_id

    @property
    def building_id(self):
        if self._building_id is None:
            if self._bd.id is not None:
                self._building_id = self._bd.id
        return self._building_id

    @property
    def sp(self):
        return self._sp

    @sp.setter
    def sp(self, obj):
        # Could add assertions here for type?
        self._sp = obj
        self._soil_profile_id = obj.id

    @property
    def fd(self):
        return self._fd

    @fd.setter
    def fd(self, obj):
        # Could add assertions here for type?
        self._fd = obj
        self._foundation_id = obj.id

    @property
    def bd(self):
        return self._bd

    @bd.setter
    def bd(self, obj):
        # Could add assertions here for type?
        self._bd = obj
        self._building_id = obj.id



__author__ = 'maximmillen'


def convert_stress_to_mass(q, width, length, gravity):
    """
    Converts a foundation stress to an equivalent mass.

    :param q: applied stress [Pa]
    :param width: foundation width [m]
    :param length: foundation length [m]
    :param gravity: applied gravitational acceleration [m/s2]
    :return:
    """
    mass = q * width * length / gravity
    return mass


def clean_float(value):
    if value is None or value == "":
        return None
    return float(value)


def collect_serial_value(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, int):
        return value
    elif hasattr(value, "to_dict"):
        return value.to_dict()
    elif hasattr(value, "__len__"):
        tolist = getattr(value, "tolist", None)
        if callable(tolist):
            value = value.tolist()
            return value
        else:
            if hasattr(value, "to_dict"):
                value = value.to_dict()
                return value
            else:
                values = []
                for item in value:
                    values.append(collect_serial_value(item))
                return values
    else:
        return value


__author__ = 'maximmillen'


pp = {}  # TODO: switch to read description from docstring, or extend the string object

# generic
pp["physical_object"] = {}
pp["physical_object"]["type"] = ["Object type", ""]
pp["physical_object"]["base_type"] = ["Object type for saving and loading", ""]
pp["physical_object"]["name"] = ["Name to identify object", ""]
pp["physical_object"]["id"] = ["Id number", ""]

# soil
pp["soil"] = {}
pp["soil"]["stype"] = ["Soil model type", ""]
pp["soil"]["g_mod"] = ["Soil shear modulus", "Pa"]
pp["soil"]["bulk_mod"] = ["Soil bulk modulus", "Pa"]
pp["soil"]["poissons_ratio"] = ["Soil Poissons ratio", "Pa"]
pp["soil"]["phi"] = ["Soil friction angle", "deg"]
pp["soil"]["cohesion"] = ["Cohesive strength of soil", "Pa"]
pp["soil"]["dilation_angle"] = ["Soil dilation angle", "deg"]
pp["soil"]["e_min"] = ["Minimum void ratio", "-"]
pp["soil"]["e_max"] = ["Maximum void ratio", "-"]
pp["soil"]["e_curr"] = ["Current void ratio", "-"]
pp["soil"]["relative_density"] = ["Relative density", "-"]
pp["soil"]["specific_gravity"] = ["Specific gravity of soil", "-"]
pp["soil"]["unit_dry_weight"] = ["Dry unit weight of soil", "N/m3"]
pp["soil"]["unit_sat_weight"] = ["Saturated unit weight of soil", "N/m3"]
pp["soil"]["saturation"] = ["Saturation ratio of soil", "-"]
pp["soil"]["plasticity_index"] = ["Plasticity index of the soil", "-"]
pp["soil"]["permeability"] = ["Permeability of soil", "-"]

pp["soil_critical"] = {}
pp["soil_critical"]["e_cr0"] = ["", ""]
pp["soil_critical"]["p_cr0"] = ["", ""]
pp["soil_critical"]["lamb_crl"] = ["", ""]

# soil profile
pp["soil_profile"] = {}
pp["soil_profile"]["gwl"] = ["Ground water level", "m"]
pp["soil_profile"]["unit_water_weight"] = ["Unit weight of water", "N/m3"]
pp["soil_profile"]["height"] = ["Total height of soil profile", "m"]
pp["soil_profile"]["layers"] = ["Layers of soil specified by id and depth to top of layer", "m"]

# foundation
pp["foundation"] = {}
pp["foundation"]["ftype"] = ["Foundation type", ""]
pp["foundation"]["width"] = ["Foundation width", "m"]
pp["foundation"]["length"] = ["Foundation length", "m"]
pp["foundation"]["depth"] = ["Foundation depth from surface", "m"]
pp["foundation"]["height"] = ["Foundation height", "m"]
pp["foundation"]["mass"] = ["Foundation mass", "kg"]
pp["foundation"]["density"] = ["Foundation mass density", "kg/m3"]

# pad foundation
pp["pad_foundation"] = {}
pp["pad_foundation"]["n_pads_l"] = ["Number of footings in length direction", ""]
pp["pad_foundation"]["n_pads_w"] = ["Number of footings in width direction", ""]
pp["pad_foundation"]["pad_length"] = ["Length of footing", "m"]
pp["pad_foundation"]["pad_width"] = ["Width of footing", "m"]

# structure
pp["structure"] = {}
pp["structure"]["h_eff"] = ["Structure effective height", "m"]
pp["structure"]["mass_eff"] = ["Structure effective mass", "T"]
pp["structure"]["t_fixed"] = ["Structure effective period", "s"]
pp["structure"]["mass_ratio"] = ["Ratio of horizontal seismic mass to vertical mass", ""]

pp["building"] = {}
pp["building"]["floor_length"] = ["Length of floor area in direction of interest", "m"]
pp["building"]["floor_width"] = ["Length of floor area out-of-plane", "m"]
pp["building"]["interstorey_heights"] = ["Distance between storeys", "m"]
pp["building"]["storey_masses"] = ["Mass of each storey", "kg"]

pp["frame_building_2D"] = {}
pp["frame_building_2D"]["beams"] = ["Array of beam sections", "m"]
pp["frame_building_2D"]["columns"] = ["Array of column sections", "m"]
pp["frame_building_2D"]["bay_lengths"] = ["Length between columns", "m"]
pp["frame_building_2D"]["beam_depths"] = ["Depth of beams", "m"]
# pp["frame_building"]["n_seismic_frames"] = ["Number of seismic frames", ""]
# pp["frame_building"]["n_gravity_frames"] = ["Number of gravity frames", ""]

# hazard
pp["seismic_hazard"] = {}
pp["seismic_hazard"]["z_factor"] = ["Hazard factor (Z)", ""]
pp["seismic_hazard"]["r_factor"] = ["Return period factor (R)", ""]
pp["seismic_hazard"]["n_factor"] = ["Near-field factor (N)", ""]
pp["seismic_hazard"]["magnitude"] = ["Magnitude", ""]
pp["seismic_hazard"]["corner_period"] = ["Spectrum corner period", "s"]
pp["seismic_hazard"]["corner_acc_factor"] = ["Spectrum corner acceleration factor", ""]


import json


def read_json_sensor_file(ffp):
    """
    Reads the sensor file and stores it as a dictionary.

    :param ffp: full file path to json file
    :return:
    """
    sensor_path = ffp
    si = json.load(open(sensor_path))
    for m_type in si:
        # Convert keys from strings to integers
        si[m_type] = {int(k): v for k, v in si[m_type].items()}
    return si


def get_all_sensor_codes(si, wild_sensor_code):
    """
    Get all sensor sensor_codes that match a wild sensor code

    :param si: dict, sensor index json dictionary
    :param wild_sensor_code: str, a sensor code with "*" for wildcards (e.g. ACCX-*-L2C-*)
    :return:
    """
    mtype_and_ory, x, y, z = wild_sensor_code.split("-")
    if mtype_and_ory == "*":
        mtypes = list(si)
    elif mtype_and_ory[-1] in "XYZ" and "ACCX" not in si:  # Need to support old sensor_file.json files.
        mtypes = [mtype_and_ory[:-1]]
    else:
        mtypes = [mtype_and_ory]

    all_sensor_codes = []
    for mtype in mtypes:
        for m_number in si[mtype]:
            if x in ["*", si[mtype][m_number]['X-CODE']] and \
                    y in ["*", si[mtype][m_number]['Y-CODE']] and \
                    z in ["*", si[mtype][m_number]['Z-CODE']]:
                cc = get_sensor_code_by_number(si, mtype, m_number)
                all_sensor_codes.append(cc)

    return all_sensor_codes


def create_motion_name(test_name, sensor_code, code_suffix=""):
    """
    Builds the full name of the file

    :param test_name: str, test name
    :param sensor_code: str, a sensor code (e.g. ACCX-UB1-L2C-M)
    :param code_suffix: str, suffix
    :return:
    """
    return "%s-%s-%s" % (test_name, sensor_code, code_suffix)


def get_sensor_code_by_number(si, mtype, sensor_number, quiet=False):
    """
    Given a sensor number, get the full sensor code (e.g. ACCX-UB1-L2C-M)

    :param si: dict, sensor index json dictionary
    :param mtype: str, sensor type
    :param sensor_number: int, number of sensor
    :param quiet: bool, if true then return None if not found
    :return: str or None, sensor_code: a sensor code (e.g. ACCX-UB1-L2C-M)
    """
    try:
        if 'Orientation' in si[mtype][sensor_number]:
            orientation = si[mtype][sensor_number]['Orientation']
        else:
            orientation = ""
        return "%s%s-%s-%s-%s" % (mtype,
                                orientation,
                                si[mtype][sensor_number]['X-CODE'],
                                si[mtype][sensor_number]['Y-CODE'],
                                si[mtype][sensor_number]['Z-CODE'])
    except KeyError:
        if quiet:
            return None
        raise


def get_mtype_and_number_from_code(si, sensor_code):
    """
    Given a sensor sensor_code, get motion type and sensor number

    :param si: dict, sensor index json dictionary
    :param sensor_code: str, a sensor code (e.g. ACCX-UB1-L2C-M)
    :return:
    """
    mtype_and_ory, x, y, z = sensor_code.split("-")
    if mtype_and_ory[-1] in "XYZ" and "ACCX" not in si:  # Need to support old sensor_file.json files.
        mtype = mtype_and_ory[:-1]
    else:
        mtype = mtype_and_ory
    for m_number in si[mtype]:
        cc = get_sensor_code_by_number(si, mtype, m_number)
        if cc == sensor_code:
            return mtype, m_number
    return None, None


def get_surface_height(si):
    for mtype in si:
        for m_number in si[mtype]:
            if si[mtype][m_number]["Y-CODE"] == "S":
                return si[mtype][m_number]["y"]
    return None


def get_depth_by_code(si, sensor_code):
    mtype, sensor_number = get_mtype_and_number_from_code(si, sensor_code)
    if sensor_number is None:
        raise KeyError("Depth not found for sensor_code: %s" % sensor_code)
    max_depth = get_surface_height(si)
    return max_depth - si[mtype][sensor_number]['y']



import json
from sfsimodels.models import soils, buildings, foundations, material, systems, abstract_models
from collections import OrderedDict
from sfsimodels import models
from sfsimodels.exceptions import deprecation, ModelError


standard_types = ["soil", "soil_profile", "foundation", "building", "section", "system", "custom_type"]


def add_to_obj(obj, dictionary, exceptions=None, verbose=0):
    """
    Cycles through a dictionary and adds the key-value pairs to an object.

    :param obj:
    :param dictionary:
    :param exceptions:
    :param verbose:
    :return:
    """
    if exceptions is None:
        exceptions = []
    for item in dictionary:
        if item in exceptions:
            continue
        if dictionary[item] is not None:
            if verbose:
                print("assign: ", item, dictionary[item])
            setattr(obj, item, dictionary[item])


def load_json(ffp, custom=None, verbose=0):
    """
    Given a json file it creates a dictionary of sfsi objects

    :param ffp: str, Full file path to json file
    :param custom: dict, used to load custom objects, {model type: custom object}
    :param verbose: int, console output
    :return: dict
    """
    data = json.load(open(ffp))
    return ecp_dict_to_objects(data, custom, verbose=verbose)


def load_json_and_meta(ffp, custom=None, verbose=0):
    data = json.load(open(ffp))
    md = {}
    for item in data:
        if item != "models":
            md[item] = data[item]
    return ecp_dict_to_objects(data, custom, verbose=verbose), md


def loads_json(p_str, custom=None, meta=False, verbose=0):
    """
    Given a json string it creates a dictionary of sfsi objects

    :param ffp: str, Full file path to json file
    :param custom: dict, used to load custom objects, {model type: custom object}
    :param meta: bool, if true then also return all ecp meta data in separate dict
    :param verbose: int, console output
    :return: dict
    """
    data = json.loads(p_str)
    if meta:
        md = {}
        for item in data:
            if item != "models":
                md[item] = data[item]
        return ecp_dict_to_objects(data, custom, verbose=verbose), md
    else:
        return ecp_dict_to_objects(data, custom, verbose=verbose)


# Deprecated name
def dicts_to_objects(data, verbose=0):
    """Deprecated. Use ecp_dict_to_objects"""
    deprecation('Deprecated, dicts_to_objects should be switched to ecp_dict_to_objects')
    ecp_dict_to_objects(data, verbose=verbose)


deprecated_types = OrderedDict([
    ("structure", "sdof"),
    ("frame_building", "building_frame"),
    ("frame_building_2D", "building_frame2D"),
    ("wall_building", "building_wall"),
    ("pad_foundation", "foundation_pad"),
    ("raft_foundation", "foundation_raft")
])


def ecp_dict_to_objects(ecp_dict, custom_map=None, verbose=0):
    """
    Given an ecp dictionary, build a dictionary of sfsi objects

    :param ecp_dict: dict, engineering consistency project dictionary
    :param custom: dict, used to load custom objects, {model type: custom object}
    :param verbose: int, console output
    :return: dict
    """
    if custom_map is None:
        custom_map = {}

    obj_map = {
        "soil-soil": soils.Soil,
        "soil-soil_critical": soils.SoilCritical,
        "soil-soil_stress_dependent": soils.SoilStressDependent,
        "soil_profile-soil_profile": soils.SoilProfile,
        "building-building": buildings.Building,
        "building-frame_building": buildings.BuildingFrame,
        # "building-frame_building_2D": buildings.BuildingFrame2D,  # deprecated type
        "building-building_frame2D": buildings.BuildingFrame2D,
        # "building-wall_building": buildings.BuildingWall,  # deprecated type
        "building-building_wall": buildings.BuildingWall,
        "building-structure": buildings.BuildingSDOF,
        "building-sdof": buildings.BuildingSDOF,
        "foundation-foundation": foundations.Foundation,
        "foundation-foundation_raft": foundations.FoundationRaft,
        "foundation-raft": foundations.FoundationRaft,  # Deprecated approach for type, remove in v1
        "foundation-foundation_pad": foundations.FoundationPad,
        "section-section": buildings.Section,
        "custom_object-custom_object": abstract_models.CustomObject,
        "system-system": systems.SoilStructureSystem,  # deprecated type
        "system-sfs": systems.SoilStructureSystem
    }

    # merge and overwrite the object map with custom maps
    for item in custom_map:
        obj_map[item] = custom_map[item]

    data_models = ecp_dict["models"]

    exception_list = ["soil_profile", "building", "system"]
    objs = OrderedDict()
    collected = set([])
    # Set base type properly
    mtypes = list(data_models)
    for mtype in mtypes:
        base_type = mtype
        if base_type[:-1] in standard_types:  # support the loading of old plural based ecp files
            base_type = base_type[:-1]
            data_models[base_type] = data_models[mtype]
            del data_models[mtype]
        for m_id in data_models[base_type]:
            data_models[base_type][m_id]["base_type"] = base_type

    for mtype in data_models:
        base_type = mtype
        if base_type in exception_list:
            continue
        collected.add(base_type)
        objs[base_type] = OrderedDict()
        for m_id in data_models[mtype]:
            obj = data_models[mtype][m_id]
            if "type" not in obj:
                obj["type"] = base_type
            try:
                obj_class = obj_map["%s-%s" % (base_type, obj["type"])]
            except KeyError:
                if obj["type"] in deprecated_types:
                    try:
                        obj_class = obj_map["%s-%s" % (base_type, deprecated_types[obj["type"]])]
                    except KeyError:
                        raise KeyError("Map for Model: '%s' index: '%s' and type: '%s' not available, "
                                       "add '%s-%s' to custom dict" % (base_type, m_id, base_type, base_type, obj["type"]))
                else:
                    raise KeyError("Map for Model: '%s' index: '%s' and type: '%s' not available, "
                                   "add '%s-%s' to custom dict" % (base_type, m_id, base_type, base_type, obj["type"]))
            new_instance = obj_class()
            add_to_obj(new_instance, data_models[mtype][m_id], verbose=verbose)
            # print(mtype, m_id)
            objs[base_type][int(data_models[mtype][m_id]["id"])] = new_instance

    # Deal with all the exceptions
    for mtype in data_models:
        base_type = mtype

        if base_type in collected:
            continue
        if base_type not in objs:
            objs[base_type] = OrderedDict()

        if base_type == "soil_profile":
            if "soil" not in objs:
                objs["soil"] = OrderedDict()
            for m_id in data_models[mtype]:
                obj = data_models[mtype][m_id]
                if "type" not in obj:
                    obj["type"] = base_type
                try:
                    obj_class = obj_map["%s-%s" % (base_type, obj["type"])]
                except KeyError:
                    raise KeyError("Map for Model: '%s' index: '%s' and type: '%s' not available, "
                                   "add '%s-%s' to custom dict" % (mtype, m_id, base_type, base_type, obj["type"]))
                new_soil_profile = obj_class()
                new_soil_profile.id = data_models[mtype][m_id]["id"]
                for j in range(len(data_models[mtype][m_id]["layers"])):
                    depth = data_models[mtype][m_id]['layers'][j]["depth"]
                    soil = objs["soil"][int(data_models[mtype][m_id]['layers'][j]["soil_id"])]
                    new_soil_profile.add_layer(depth, soil)
                add_to_obj(new_soil_profile, data_models[mtype][m_id], exceptions=["layers"], verbose=verbose)
                objs[base_type][int(data_models[mtype][m_id]["id"])] = new_soil_profile

        if base_type == "building":
            for m_id in data_models[mtype]:
                obj = data_models[mtype][m_id]
                if obj["type"] in deprecated_types:
                    obj["type"] = deprecated_types[obj["type"]]
                if "type" not in obj:
                    obj["type"] = base_type
                if obj["type"] in ["sdof"]:
                    new_building = obj_map["%s-%s" % (base_type, obj["type"])]()
                elif "building_frame" in obj["type"]:
                    n_storeys = len(obj['interstorey_heights'])
                    n_bays = len(obj['bay_lengths'])
                    new_building = obj_map["%s-%s" % (base_type, obj["type"])](n_storeys, n_bays)
                    for ss in range(n_storeys):
                        for bb in range(n_bays):
                            sect_is = obj["beam_section_ids"][ss][bb]
                            if hasattr(sect_is, "__len__"):
                                n_sections = len(sect_is)
                                new_building.beams[ss][bb].split_into_multiple([1] * n_sections)  # TODO: should be lengths
                                for sect_i in range(len(sect_is)):
                                    beam_sect_id = str(obj["beam_section_ids"][ss][bb][sect_i])
                                    sect_dictionary = obj["beam_sections"][beam_sect_id]
                                    add_to_obj(new_building.beams[ss][bb].sections[sect_i], sect_dictionary, verbose=verbose)
                            else:  # deprecated loading
                                beam_sect_id = str(obj["beam_section_ids"][ss][bb])
                                sect_dictionary = obj["beam_sections"][beam_sect_id]
                                add_to_obj(new_building.beams[ss][bb].section[0], sect_dictionary, verbose=verbose)
                        for cc in range(n_bays + 1):
                            sect_is = obj["column_section_ids"][ss][cc]
                            if hasattr(sect_is, "__len__"):
                                n_sections = len(sect_is)
                                # TODO: should be lengths
                                new_building.columns[ss][cc].split_into_multiple([1] * n_sections)
                                for sect_i in range(len(sect_is)):
                                    column_sect_id = str(obj["column_section_ids"][ss][cc][sect_i])
                                    sect_dictionary = obj["column_sections"][column_sect_id]
                                    add_to_obj(new_building.columns[ss][cc].sections[sect_i], sect_dictionary, verbose=verbose)
                            else:
                                column_sect_id = str(obj["column_section_ids"][ss][cc])
                                sect_dictionary = obj["column_sections"][column_sect_id]
                                add_to_obj(new_building.columns[ss][cc].section[0], sect_dictionary, verbose=verbose)

                else:
                    n_storeys = len(obj['interstorey_heights'])
                    new_building = obj_map["%s-%s" % (base_type, obj["type"])](n_storeys)
                add_to_obj(new_building, data_models[mtype][m_id], verbose=verbose)
                objs[base_type][int(data_models[mtype][m_id]["id"])] = new_building

        if base_type == "system":  # must be run after other objects are loaded
            for m_id in data_models[mtype]:
                obj = data_models[mtype][m_id]
                if "type" not in obj:
                    obj["type"] = base_type
                try:
                    obj_class = obj_map["%s-%s" % (base_type, obj["type"])]
                except KeyError:
                    raise KeyError("Map for Model: '%s' index: '%s' and type: '%s' not available, "
                                   "add '%s-%s' to custom dict" % (mtype, m_id, base_type, base_type, obj["type"]))
                new_system = obj_class()

                # Attach the soil profile
                soil_profile_id = data_models[mtype][m_id]['soil_profile_id']
                soil_profile = objs["soil_profile"][int(soil_profile_id)]
                new_system.sp = soil_profile

                # Attach the foundation
                foundation_id = data_models[mtype][m_id]['foundation_id']
                foundation = objs["foundation"][int(foundation_id)]
                new_system.fd = foundation

                # Attach the building
                building_id = data_models[mtype][m_id]['building_id']
                building = objs["building"][int(building_id)]
                new_system.bd = building

                # Add remaining parameters
                ignore_list = ["foundation_id", "building_id", "soil_profile_id"]
                add_to_obj(new_system, data_models[mtype][m_id], exceptions=ignore_list, verbose=verbose)
                objs[base_type][int(data_models[mtype][m_id]["id"])] = new_system

    all_bts = list(objs)
    for base_type in all_bts:  # Support for old style ecp file
        if base_type in standard_types:
            objs[base_type + "s"] = objs[base_type]
    return objs


class Output(object):
    name = ""
    units = ""
    doi = ""
    sfsimodels_version = ""
    comments = ""

    def __init__(self):
        self.unordered_models = OrderedDict()

    def add_to_dict(self, an_object, extras=None):
        """
        Convert models to json serialisable output

        :param an_object: An instance of a model object
        :param extras: A dictionary of extra variables that should be
        :return:
        """
        if an_object.id is None:
            raise ModelError("id must be set on object before adding to output.")
        if hasattr(an_object, "base_type"):
            mtype = an_object.base_type
        elif hasattr(an_object, "type"):
            if an_object.type in standard_types:
                mtype = an_object.type
            else:
                mtype = "custom_type"
        else:
            raise ModelError("Object does not have attribute 'base_type' or 'type', cannot add to output.")
        if mtype not in self.unordered_models:  # Catch any custom objects
            self.unordered_models[mtype] = OrderedDict()
        if mtype == "soil_profile":
            if "soil" not in self.unordered_models:
                self.unordered_models["soil"] = OrderedDict()
            profile_dict = an_object.to_dict()
            profile_dict["layers"] = []
            for layer in an_object.layers:
                self.unordered_models["soil"][an_object.layers[layer].id] = an_object.layers[layer].to_dict()
                profile_dict["layers"].append({
                    "soil_id": str(an_object.layers[layer].id),
                    "depth": float(layer)
                })

            self.unordered_models["soil_profile"][an_object.id] = profile_dict
        elif hasattr(an_object, "to_dict"):
            self.unordered_models[mtype][an_object.id] = an_object.to_dict()
        else:
            raise ModelError("Object does not have method 'to_dict', cannot add to output.")

    def add_to_output(self, mtype, m_id, serialisable_dict):
        """
        Can add additional objects or dictionaries to output file that don't conform to standard objects.

        :param mtype:
        :param m_id:
        :param serialisable_dict:
        :return:
        """
        if mtype not in self.unordered_models:
            self.unordered_models[mtype] = OrderedDict()
        self.unordered_models[mtype][m_id] = serialisable_dict

    @property
    def models(self):
        models_dict = OrderedDict()
        collected = []
        for item in standard_types:
            if item in self.unordered_models:
                models_dict[item] = self.unordered_models[item]
                collected.append(item)
        for item in self.unordered_models:
            if item not in collected:
                models_dict[item] = self.unordered_models[item]
        return models_dict

    @staticmethod
    def parameters():
        return ["name", "units", "doi", "sfsimodels_version", "comments", "models"]

    def to_dict(self):
        outputs = OrderedDict()
        for item in self.parameters():
            outputs[item] = self.__getattribute__(item)
        return outputs



from collections import OrderedDict

import numpy as np

from sfsimodels.exceptions import ModelError
from sfsimodels.models.abstract_models import PhysicalObject
from sfsimodels import checking_tools as ct
from sfsimodels.exceptions import deprecation


class Foundation(PhysicalObject):
    """
    An object to describe building foundations
    """
    _id = None
    name = None
    _width = None  # [m], The length of the foundation in the direction of shaking
    _length = None  # [m], The length of the foundation perpendicular to the shaking
    _depth = None  # [m], The depth of the foundation from the surface
    _height = None  # [m], The height of the foundation from base of foundation to ground floor
    _density = None  # [kg/m3], Density of foundation
    _mass = None  # kg
    base_type = "foundation"
    type = "foundation"
    _tolerance = 0.0001  # consistency tolerance

    _extra_class_inputs = [
        "id",
        "name",
        "base_type",
        "type",
        "width",
        "length",
        "depth",
        "height",
        "density",
        "mass"
    ]

    def __str__(self):
        return "Foundation id: {0}, name: {1}".format(self.id, self.name)

    def __format__(self, format_spec):
        return "Foundation"

    def __init__(self):
        super(Foundation, self).__init__()
        self.inputs = [
        "id",
        "name",
        "base_type",
        "type",
        "width",
        "length",
        "depth",
        "height",
        "density",
        "mass"
    ]

    def to_dict(self):
        outputs = OrderedDict()
        skip_list = []
        for item in self.inputs:
            if item not in skip_list:
                value = self.__getattribute__(item)
                if isinstance(value, int):
                    outputs[item] = str(value)
                else:
                    outputs[item] = value
        return outputs

    @property
    def ancestor_types(self):
        return super(Foundation, self).ancestor_types + ["foundation"]

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def area(self):
        """
        Foundation area in plan
        :return:
        """
        try:
            return self.length * self.width
        except TypeError:
            return None

    @property
    def length(self):
        """
        Length of the foundation (typically in the out-of-plane direction)
        :return:
        """
        return self._length

    @property
    def width(self):
        """
        Length of the foundation (typically in the in-plane direction)
        :return:
        """
        return self._width

    @property
    def height(self):
        """
        Measure of the base of the foundation to the top
        :return:
        """
        return self._height

    @property
    def depth(self):
        """
        Measure of the base of the foundation to the surface of the soil
        :return:
        """
        return self._depth

    @property
    def mass(self):
        """
        The mass of the whole foundation
        :return:
        """
        return self._mass

    @property
    def density(self):
        """
        The mass density of the foundation [kg/m3]
        :return:
        """
        return self._density

    @property
    def weight(self):
        """
        The weight of the foundation [N]
        :return:
        """
        return self.mass * 9.8

    @length.setter
    def length(self, value):
        if value is None or value == "":
            return
        self._length = float(value)

    @width.setter
    def width(self, value):
        if value is None or value == "":
            return
        self._width = float(value)

    @height.setter
    def height(self, value):
        if value is None or value == "":
            return
        self._height = float(value)

    @depth.setter
    def depth(self, value):
        if value is None or value == "":
            return
        self._depth = float(value)

    @density.setter
    def density(self, value, override=False):
        if value is None or value == "":
            return
        density = self._calc_density()
        if density is not None and not np.isclose(density, value, rtol=self._tolerance) and not override:
            raise ModelError("Density inconsistent with set mass")
        self._density = float(value)
        mass = self._calc_mass()
        if mass is not None and not ct.isclose(mass, self.mass):
            self.mass = mass

    @mass.setter
    def mass(self, value, override=False):
        if value is None or value == "":
            return
        mass = self._calc_mass()
        if mass is not None and not ct.isclose(mass, value, rel_tol=self._tolerance) and not override:
            raise ModelError("Mass inconsistent with set density")
        self._mass = float(value)
        density = self._calc_density()
        if density is not None and not ct.isclose(density, self.density, rel_tol=self._tolerance):
            self.density = density

    def _calc_mass(self):
        try:
            return self.area * self.height * self.density
        except TypeError:
            return None

    def _calc_density(self):
        try:
            return self.mass / (self.area * self.height)
        except TypeError:
            return None
        except ZeroDivisionError:
            return None


class FoundationRaft(Foundation):
    """
    An extension to the Foundation Object to describe Raft foundations
    """
    ftype = "raft"
    type = "foundation_raft"
    _extra_class_inputs = []

    def __str__(self):
        return "FoundationRaft id: {0}, name: {1}".format(self.id, self.name)

    def __init__(self):
        super(FoundationRaft, self).__init__()
        self.inputs = self.inputs + self._extra_class_inputs

    @property
    def ancestor_types(self):
        return super(FoundationRaft, self).ancestor_types + ["foundation_raft"]

    @property
    def i_ww(self):
        """
        Contact moment-area around the width axis
        :return:
        """
        return self.width * self.length ** 3 / 12

    @property
    def i_ll(self):
        """
        Contact moment-area around the length axis
        :return:
        """
        return self.length * self.width ** 3 / 12


class FoundationPad(Foundation):
    """
    An extension to the Foundation Object to describe Pad foundations
    """
    ftype = "pad"
    type = "foundation_pad"
    n_pads_l = 4  # Number of pads in length direction
    n_pads_w = 3  # Number of pads in width direction
    pad_length = 1.0  # m  # TODO: make parameters protected
    pad_width = 1.0  # m
    _extra_class_inputs = [
        "n_pads_l",
        "n_pads_w",
        "pad_length",
        "pad_width",
    ]

    def __str__(self):
        return "PadFoundation id: {0}, name: {1}".format(self.id, self.name)

    def __init__(self):
        super(FoundationPad, self).__init__()
        self.inputs += self._extra_class_inputs

    @property
    def ancestor_types(self):
        return super(FoundationPad, self).ancestor_types + ["foundation_pad"]

    @property
    def i_ww(self):
        """
        Second moment of inertia around the width axis.
        :return:
        """
        d_values = []
        for i in range(self.n_pads_l):
            d_values.append(self.pad_position_l(i))
        d_values = np.array(d_values) - self.length / 2
        area_d_sqrd = sum(self.pad_area * d_values ** 2) * self.n_pads_w
        i_second = self.pad_i_ww * self.n_pads
        return area_d_sqrd + i_second

    @property
    def i_ll(self):
        """
        Second moment of inertia around the length axis.
        :return:
        """
        d_values = []
        for i in range(self.n_pads_w):
            d_values.append(self.pad_position_w(i))
        d_values = np.array(d_values) - self.width / 2
        area_d_sqrd = sum(self.pad_area * d_values ** 2) * self.n_pads_l
        i_second = self.pad_i_ll * self.n_pads
        return area_d_sqrd + i_second

    @property
    def n_pads(self):
        """
        Total number of pad footings
        :return:
        """
        return self.n_pads_w * self.n_pads_l

    @property
    def pad_area(self):
        """
        Area of a pad
        :return:
        """
        return self.pad_length * self.pad_width

    @property
    def pad_i_ww(self):
        """
        Second moment of inertia of a single pad around the width axis.
        :return:
        """
        return self.pad_length ** 3 * self.pad_width / 12

    @property
    def pad_i_ll(self):
        """
        Second moment of inertia of a single pad around the length axis.
        :return:
        """
        return self.pad_width ** 3 * self.pad_length / 12

    @property
    def area(self):
        """
        Contact area of the whole foundation in plan
        :return:
        """
        return self.n_pads * self.pad_area

    def pad_position_l(self, i):
        """
        Determines the position of the ith pad in the length direction.
        Assumes equally spaced pads.

        :param i: ith number of pad in length direction (0-indexed)
        :return:
        """
        if i >= self.n_pads_l:
            raise ModelError("pad index out-of-bounds")
        return (self.length - self.pad_length) / (self.n_pads_l - 1) * i + self.pad_length / 2

    def pad_position_w(self, i):
        """
        Determines the position of the ith pad in the width direction.
        Assumes equally spaced pads.

        :param i: ith number of pad in width direction (0-indexed)
        :return:
        """
        if i >= self.n_pads_w:
            raise ModelError("pad index out-of-bounds")
        return (self.width - self.pad_width) / (self.n_pads_w - 1) * i + self.pad_width / 2


class PadFoundation(FoundationPad):
    def __init__(self):
        deprecation("PadFoundation class is deprecated, use FoundationPad.")
        super(PadFoundation, self).__init__()


class RaftFoundation(FoundationRaft):
    def __init__(self):
        deprecation("RaftFoundation class is deprecated, use FoundationRaft.")
        super(RaftFoundation, self).__init__()



from sfsimodels.models.abstract_models import PhysicalObject, CustomObject
from sfsimodels.models.hazards import SeismicHazard
from sfsimodels.models.foundations import Foundation, FoundationPad, FoundationRaft
from sfsimodels.models.soils import Soil, SoilCritical, SoilProfile, SoilStressDependent
from sfsimodels.models.buildings import Building, BuildingFrame, BuildingWall, BuildingSDOF, BuildingFrame2D, Section

from sfsimodels.models import material
from sfsimodels.models.systems import SoilStructureSystem
from sfsimodels.models.time import TimeSeries
from sfsimodels.output import format_value, format_name, output_to_table
from sfsimodels.files import ecp_dict_to_objects, load_json, loads_json, add_to_obj, Output
from sfsimodels.functions import clean_float, collect_serial_value
from sfsimodels.exceptions import DesignError, AnalysisError, ModelError, ModelWarning
from sfsimodels import sensors

# from sfsimodels.build_model_descriptions import print_all_parameters
from sfsimodels.models.foundations import PadFoundation, RaftFoundation  # deprecated objects
from sfsimodels.models.buildings import FrameBuilding, WallBuilding, Structure, FrameBuilding2D  # deprecated objects


import warnings


class DesignError(Exception):
    pass  # deprecated


class AnalysisError(Exception):
    pass


class ModelError(Exception):
    pass


class ModelWarning(Warning):
    pass


def deprecation(message):
    warnings.warn(message, stacklevel=3)
